/////////////////////////////////////////////////////////////////////////////
// Name:        ax_utils.cpp
// Purpose:     AxTk framework - utilities for archives and more
// Author:      Julian Smart
// Modified by: 
// Created:     2009-02-27
// RCS-ID:      
// Copyright:   (c) Julian Smart
// Licence:     New BSD License
/////////////////////////////////////////////////////////////////////////////

#include "wx/wx.h"

#include "wx/filename.h"
#include "wx/textfile.h"
#include "wx/txtstrm.h"
#include "wx/datstrm.h"
#include "wx/wfstream.h"
#include "wx/dir.h"
#include "wx/utils.h"
#include "wx/zipstrm.h"
#include "wx/zstream.h"
#include "wx/tarstrm.h"
#include "wx/progdlg.h"
#include "wx/tokenzr.h"
#include "wx/stdpaths.h"
#include "wx/filedlg.h"
#include "wx/dirdlg.h"
#include "wx/mimetype.h"
#include "wx/notebook.h"
#include "wx/process.h"

#ifdef __WXMSW__
#include <windows.h>
#include "wx/msw/winundef.h"
#include "wx/msw/uxtheme.h"
#include "wx/msw/private.h"
#endif

#ifdef __WXMAC__
#include "wx/mac/private.h"
#ifndef __DARWIN__
#include "InternetConfig.h"
#endif

#ifdef __DARWIN__
extern "C" {
#include <mach/mach.h>
}
#include <CoreFoundation/CFMachPort.h>
#endif

#endif

#ifdef __UNIX__
// For access
#include <unistd.h>
#include  <sys/stat.h>
#endif

#include "ax/ax_utils.h"

/*
 * AxFileUtilities
 */

// Get a suitable directory for putting documents by default
wxString AxFileUtilities::GetDocumentDirectory()
{
    wxString docDir;
#ifdef __WXMSW__
    wxStandardPaths stdPaths;
    docDir = stdPaths.GetDocumentsDir();
    if (!docDir.IsEmpty())
        return docDir;

    docDir = wxGetenv(wxT("USERPROFILE"));
    if (!docDir.IsEmpty())
    {
        wxString d(docDir + wxT("\\Documents"));
        if (wxDirExists(d))
            docDir = d;
        else
        {
            d = docDir + wxT("\\My Documents");
            if (wxDirExists(d))
                docDir = d;
        }
    }
#endif

#if defined(__UNIX__) && !defined(__WXMAC__)
    if (docDir.IsEmpty() && !wxDirExists(wxGetHomeDir() + wxT("/My Documents")) && !wxDirExists(wxGetHomeDir() + wxT("/Documents")))
    {
        wxStandardPaths stdPaths;
        wxString stdDocDir = stdPaths.GetDocumentsDir();
        if (stdDocDir != wxGetHomeDir())
            docDir = stdDocDir;
    }
#endif

    if (docDir.IsEmpty())
    {
        docDir = wxGetHomeDir() + wxT("/My Documents");
        if (wxDirExists(docDir))
        {
            // Do nothing
        }
        else
        {
            docDir = wxGetHomeDir() + wxT("/Documents");
            if (wxDirExists(docDir))
            {
                // Do nothing
            }
            else
            {
                if (!wxDirExists(docDir)) // use DirExists in case it's a symbol link
                    if (!wxMkdir(docDir))
                        docDir = wxGetHomeDir();
            }
        }
    }
    return docDir;
}

// Get the application data directory for the current user
wxString AxFileUtilities::GetUserAppDataDirectory(const wxString& appName)
{
    wxUnusedVar(appName);

    wxStandardPaths paths;
    wxString dataDir = paths.GetUserDataDir();
    if (dataDir.IsEmpty())
        dataDir = GetDocumentDirectory();

#if defined(__WXMAC__)
    dataDir = wxMacFindFolder(  (short) kUserDomain, kApplicationSupportFolderType, kDontCreateFolder ) ;
    dataDir += wxTheApp->GetAppName();
#elif defined(__UNIX__)
    // Avoid conflict with the dotfile used for general config data
    dataDir += wxT("Data");
    dataDir.Replace(wxT(" "), wxT(""));
    dataDir.Replace(wxT("'"), wxT(""));
#endif

    if (!wxDirExists(dataDir))
        wxMkdir(dataDir);

    return dataDir;
}

// Append paths without duplicating separator
wxString AxFileUtilities::AppendPaths(const wxString& p1, const wxString& p2, const wxString& sep)
{
    wxString path = p1;

    if (p1.Last() != wxT('\\') && p1.Last() != wxT('/') && p2[0] != wxT('\\') && p2[0] != wxT('/'))
        path += sep;

    path += p2;

    return path;
}


// Change the Unix file mode
bool AxFileUtilities::Chmod(const wxString& path, int mode, bool addMode)
{
#ifdef __UNIX__
    int m = mode;
    if (addMode)
    {
        m |= GetFileMode(path);
    }
    if ( chmod( (const char*) path.fn_str(), mode) == -1 )
    {
        wxLogSysError(_("Failed to set file permission for %s"), path.c_str());
        return false;
    }
#else
    wxUnusedVar(path);
    wxUnusedVar(mode);
    wxUnusedVar(addMode);
#endif
    return true;
}

// Find the given file under the a path and change its mode
bool AxFileUtilities::FindAndChmod(const wxString& searchPath, const wxString& filename, int mode, bool addMode)
{
    bool success = false;
    wxArrayString files;
    wxDir::GetAllFiles(searchPath, & files);
    size_t i;
    for (i = 0; i < files.GetCount(); i++)
    {
        wxString fullPath(files[i]);
        // wxPrintf(wxT("%s\n"), fullPath.c_str());
        if (wxFileNameFromPath(fullPath) == filename && wxFileExists(fullPath))
        {
            if (Chmod(fullPath, mode, addMode))
                success = true;
        }
    }
    return success;
}

// Get the Unix file mode. If path is empty, get the default mode for a new file.
int AxFileUtilities::GetFileMode(const wxString& path)
{
#ifdef __UNIX__
    // the temp file should have the same permissions as the original one
    mode_t mode;

    wxStructStat st;
    if ( !path.IsEmpty() && stat( (const char*) path.fn_str(), &st) == 0 )
    {
        mode = st.st_mode;
    }
    else
    {
        // file probably didn't exist, just give it the default mode _using_
        // user's umask (new files creation should respect umask)
        mode_t mask = umask(0777);
        mode = 0666 & ~mask;
        umask(mask); // sets the default create mode for the current process. See http://en.wikipedia.org/wiki/Umask
    }

    return mode;

#else
    wxUnusedVar(path);

    return 0;
#endif
}


// Make sure the folder exists, otherwise create recursively.
bool AxFileUtilities::MkDirRecursively(const wxString& path)
{
    if (wxDirExists(path))
        return true;

    wxString p;

    wxStringTokenizer tkz(path, wxT("/\\"));
    p = tkz.GetNextToken();

    while (tkz.HasMoreTokens())
    {
        wxString token = tkz.GetNextToken();

        p += wxFILE_SEP_PATH + token;
        if (!wxDirExists(p))
        {
            if (!wxMkdir(p))
                return false;
        }
    }
    return true;
}

// Remove dir recursively
bool AxFileUtilities::RemoveDirRecursively(const wxString& path)
{
    // First get any directories, and recursively remove them.
    wxArrayString dirs;
        
    {
        wxString d;
        wxDir dir;
        if (!dir.Open(path))
            return false;
        
        bool success = dir.GetFirst(& d, wxT("*"), wxDIR_DIRS);
        while (success)
        {
            wxString s = AxFileUtilities::AppendPaths(path, d);
            dirs.Add(s);
            
            success = dir.GetNext(& d);
        }
    }

    size_t i;
    for (i = 0; i < dirs.GetCount(); i++)
    {
        wxString d = dirs[i];
        if (wxDirExists(d) && d != wxT(".") && d != wxT(".."))
        {
            RemoveDirRecursively(d);
        }
    }

    wxArrayString paths;
    wxDir::GetAllFiles(path, & paths, wxEmptyString, wxDIR_FILES);
    for (i = 0; i < paths.GetCount(); i++)
    {
        wxString p = paths[i];
        wxRemoveFile(p);
    }

    wxRmdir(path);

    return true;
}

// Copy a fixed size of data from an input stream to an output stream,
// using a buffer to speed it up.
bool AxFileUtilities::BufferedCopy(wxInputStream& inStream, wxOutputStream& outStream, size_t size)
{
    static unsigned char buf[1024];
    size_t bytesLeft = size;

    while (bytesLeft > 0)
    {
        size_t bytesToRead = wxMin((size_t) 1024, bytesLeft);

        inStream.Read((void*) buf, bytesToRead);
        outStream.Write((void*) buf, bytesToRead);

        bytesLeft -= bytesToRead;
    }

    return TRUE;
}

// Copy data from an input stream to an output stream, using a buffer to speed it up.
// This version doesn't need a size.
bool AxFileUtilities::BufferedCopy(wxInputStream& inStream, wxOutputStream& outStream)
{
    static unsigned char buf[1024];

    while (!inStream.Eof())
    {
        inStream.Read((void*) buf, 1024);
        int lastRead = inStream.LastRead();
        if (lastRead > 0)
            outStream.Write((void*) buf, lastRead);
    }

    return TRUE;
}

// Copy file to stream
bool AxFileUtilities::CopyFileToStream(wxOutputStream& stream, const wxString& filename)
{
    wxFile file;
    if (!file.Open(filename))
        return FALSE;

    size_t size = (size_t) file.Length();
    file.Close();

    wxFileInputStream inStream(filename);
    if (!inStream.Ok())
        return FALSE;

    return BufferedCopy(inStream, stream, size);
}

// Copy stream to new file
bool AxFileUtilities::CopyStreamToFile(wxInputStream& stream, const wxString& filename, size_t size)
{
    wxFileOutputStream outStream(filename);
    if (!outStream.Ok())
        return FALSE;

    return BufferedCopy(stream, outStream, size);
}

// Remove trailing separator
bool AxFileUtilities::RemoveTrailingSeparator(wxString& filename)
{
    if (filename.Length() > 0 && (filename[filename.Length()-1] == wxT('/') || filename[filename.Length()-1] == wxT('\\')))
    {
        filename = filename.Mid(0, filename.Length()-1);
    }
    return true;
}

// Does the filename have this extension?
bool AxFileUtilities::HasExtension(const wxString& filename, const wxString& ext)
{
    wxString file, name, ext2;
    wxSplitPath(filename, & file, & name, & ext2);
    return (ext2.Lower() == ext.Lower());
}

// Gets the extension
wxString AxFileUtilities::GetExtension(const wxString& filename)
{
    wxString file, name, ext;
    wxSplitPath(filename, & file, & name, & ext);
    return ext;
}

// Ensure extension is present
void AxFileUtilities::EnsureExtension(wxString& path, const wxString& requiredExt)
{
    if (path.IsEmpty())
        return;

    wxString ext = GetExtension(path);
    ext.MakeLower();
    if (ext.IsEmpty() || ext != requiredExt)
    {
        StripExtension(path);
        path += wxT(".");
        path += requiredExt;
    }
}

// Intelligent strip extension: don't strip before a separator,
// and don't strip if the 'extension' is different from the passed extension
// that we expect - can be comma-separated list.
bool AxFileUtilities::StripExtension(wxString& path, const wxString& validExtensions)
{
    wxString p, name, ext;
    wxSplitPath(path, &p, &name, &ext);

    if (!validExtensions.IsEmpty() && !ext.IsEmpty() && validExtensions.Find(ext) == wxNOT_FOUND)
        return false;
    else if (!ext.IsEmpty())
    {
        path = path.Left(path.Length() - ext.Length() - 1);
        return true;
    }
    else
        return false;
}

bool AxFileUtilities::SplitPath(const wxString& fullpath,
                           wxString *pstrPath,
                           wxString *pstrName,
                           wxString *pstrExt,
                           wxPathFormat format)
{
    wxFileName::SplitPath(fullpath, pstrPath, pstrName, pstrExt, format);

    if (pstrExt)
    {
        if (pstrExt->Find(wxT("\"")) != wxNOT_FOUND ||
            pstrExt->Find(wxT(" ")) != wxNOT_FOUND ||
            pstrExt->Find(wxT("-")) != wxNOT_FOUND)
        {
            *pstrExt = wxEmptyString;
        }
    }

    return true;
}


// translate the filenames before passing them to OS functions
#define OS_FILENAME(s) (s.fn_str())

// Replacement for wxWidgets function that fails if can't set permissions
bool AxFileUtilities::CopyFile(const wxString& file1, const wxString& file2, int flags)
{
    bool overwrite = (flags & AX_COPYFILE_OVERWRITE) != 0;
    bool ignorePermissions = (flags & AX_COPYFILE_IGNORE_PERMISSIONS) != 0;
    wxUnusedVar(ignorePermissions);

#if defined(__WIN32__) && !defined(__WXMICROWIN__)
    // CopyFile() copies file attributes and modification time too, so use it
    // instead of our code if available
    //
    // NB: 3rd parameter is bFailIfExists i.e. the inverse of overwrite
    if ( !::CopyFile(file1, file2, !overwrite) )
    {
        wxLogSysError(_("Failed to copy the file '%s' to '%s'"),
                      file1.c_str(), file2.c_str());

        return false;
    }
#elif defined(__OS2__)
    if ( ::DosCopy((PSZ)file1.c_str(), (PSZ)file2.c_str(), overwrite ? DCPY_EXISTING : 0) != 0 )
        return false;
#elif defined(__PALMOS__)
    // TODO with http://www.palmos.com/dev/support/docs/protein_books/Memory_Databases_Files/
    return false;
#elif wxUSE_FILE // !Win32

    wxStructStat fbuf;

    // get permissions of file1
    bool permissionsSucceeded = (wxStat( file1.c_str(), &fbuf) == 0);

    if ( !ignorePermissions && !permissionsSucceeded)
    {
        // the file probably doesn't exist or we haven't the rights to read
        // from it anyhow
        wxLogSysError(_("Impossible to get permissions for file '%s'"),
                      file1.c_str());
        return false;
    }

    // open file1 for reading
    wxFile fileIn(file1, wxFile::read);
    if ( !fileIn.IsOpened() )
        return false;

    // remove file2, if it exists. This is needed for creating
    // file2 with the correct permissions in the next step
    if ( wxFileExists(file2)  && (!overwrite || !wxRemoveFile(file2)))
    {
        wxLogSysError(_("Impossible to overwrite the file '%s'"),
                      file2.c_str());
        return false;
    }

    // reset the umask as we want to create the file with exactly the same
    // permissions as the original one
    wxCHANGE_UMASK(0);

    // create file2 with the same permissions than file1 and open it for
    // writing

    wxFile fileOut;
    if ( !fileOut.Create(file2, overwrite, fbuf.st_mode & 0777) )
        return false;

    // copy contents of file1 to file2
    char buf[4096];
    size_t count;
    for ( ;; )
    {
        count = fileIn.Read(buf, WXSIZEOF(buf));
        if ( fileIn.Error() )
            return false;

        // end of file?
        if ( !count )
            break;

        if ( fileOut.Write(buf, count) < count )
            return false;
    }

    // we can expect fileIn to be closed successfully, but we should ensure
    // that fileOut was closed as some write errors (disk full) might not be
    // detected before doing this
    if ( !fileIn.Close() || !fileOut.Close() )
        return false;

#if !defined(__VISAGECPP__) && !defined(__WXMAC__) || defined(__UNIX__)
    // no chmod in VA.  Should be some permission API for HPFS386 partitions
    // however
    bool changedPermissions = false;
    if (permissionsSucceeded)
        changedPermissions = (chmod(OS_FILENAME(file2), fbuf.st_mode) == 0 );

    if (!changedPermissions && !ignorePermissions)
    {
        wxLogSysError(_("Impossible to set permissions for the file '%s'"),
            file2.c_str());
        return false;
    }
#endif // OS/2 || Mac

#else // !Win32 && ! wxUSE_FILE

    // impossible to simulate with wxWidgets API
    wxUnusedVar(file1);
    wxUnusedVar(file2);
    wxUnusedVar(overwrite);
    return false;

#endif // __WXMSW__ && __WIN32__

    return true;
}

// Copy files in sourceDir to destDir
bool AxFileUtilities::CopyDir(const wxString& sourceDir, const wxString& destDir, bool ifNotExists, const wxString& wildcard, int dirFlags)
{
    if (!wxDirExists(destDir))
        wxMkdir(destDir);

    bool success = true;

    if (wxDirExists(sourceDir))
    {
        wxDir dir;

        // First just get the files
        wxArrayString sourceFiles;
        dir.GetAllFiles(sourceDir, & sourceFiles, wildcard, dirFlags & ~wxDIR_DIRS);

        size_t i;
        for (i = 0; i < sourceFiles.GetCount(); i++)
        {
            wxString sourceFilename = sourceFiles[i];
            wxString destFilename = destDir + wxFILE_SEP_PATH + wxFileNameFromPath(sourceFilename);

            if (!ifNotExists || !wxFileExists(destFilename))
            {
                if (dirFlags & wxDIR_DIRS)
                {
                    // Make dest dir if necessary
                    wxString p(wxPathOnly(destFilename));
                    if (!wxDirExists(p))
                        MkDirRecursively(p);
                }

                if (!CopyFile(sourceFilename, destFilename))
                    success = false;
            }
        }

        // Now recurse
        if (dirFlags & wxDIR_DIRS)
        {
            wxArrayString sourceDirs;
            wxDir dirs;

            wxString dirName;
            if (dirs.Open(sourceDir))
            {
                bool cont = dirs.GetFirst(&dirName, wxT("*"), dirFlags & ~wxDIR_FILES);
                while ( cont )
                {
                    sourceDirs.Add(dirName);

                    cont = dirs.GetNext(&dirName);
                }
            }
            
            for (i = 0; i < sourceDirs.GetCount(); i++)
            {
                wxString newSourceDir = sourceDir + wxFILE_SEP_PATH + sourceDirs[i];
                wxString newDestDir = destDir + wxFILE_SEP_PATH + sourceDirs[i];

                success = CopyDir(newSourceDir, newDestDir, ifNotExists, wildcard, dirFlags);
            }            
        }
    }

    return success;
}

#ifdef __WXMAC__

/*
Below FinderLaunch function comes from:
http://developer.apple.com/technotes/tn/tn1002.html#fndrask
*/
    /* FinderLaunch converts a list of nTargets FSSpec records
    pointed to by the targetList parameter and converts the
    list to an Apple Event.  It then sends that event to the
    Finder.  The array of FSSpec records pointed to by the
    targetList parameter may contain references to files,
    folders, or applications.  The net effect of this command
    is equivalent to the user selecting an icon in one of the
    Finder's windows and then choosing the open command from
    the Finder's file menu. */
static OSErr FinderLaunch(long nTargets, FSSpec *targetList) {
    OSErr err;
    AppleEvent theAEvent, theReply;
    AEAddressDesc fndrAddress;
    AEDescList targetListDesc;
    OSType fndrCreator;
    Boolean wasChanged;
    AliasHandle targetAlias;
    long index;

        /* set up locals  */
    AECreateDesc(typeNull, NULL, 0, &theAEvent);
    AECreateDesc(typeNull, NULL, 0, &fndrAddress);
    AECreateDesc(typeNull, NULL, 0, &theReply);
    AECreateDesc(typeNull, NULL, 0, &targetListDesc);
    targetAlias = NULL;
    fndrCreator = 'MACS';

        /* verify parameters */
    if ((nTargets == 0) || (targetList == NULL)) {
        err = paramErr;
        goto bail;
    }

        /* create an open documents event targeting the
        finder */
    err = AECreateDesc(typeApplSignature, (Ptr) &fndrCreator,
        sizeof(fndrCreator), &fndrAddress);
    if (err != noErr) goto bail;
    err = AECreateAppleEvent(kCoreEventClass, kAEOpenDocuments,
        &fndrAddress, kAutoGenerateReturnID,
        kAnyTransactionID, &theAEvent);
    if (err != noErr) goto bail;

        /* create the list of files to open */
    err = AECreateList(NULL, 0, false, &targetListDesc);
    if (err != noErr) goto bail;
    for ( index=0; index < nTargets; index++) {
        if (targetAlias == NULL)
            err = NewAlias(NULL, (targetList + index),
                  &targetAlias);
        else err = UpdateAlias(NULL, (targetList + index),
                   targetAlias, &wasChanged);
        if (err != noErr) goto bail;
        HLock((Handle) targetAlias);
        err = AEPutPtr(&targetListDesc, (index + 1),
              typeAlias, *targetAlias,
              GetHandleSize((Handle) targetAlias));
        HUnlock((Handle) targetAlias);
        if (err != noErr) goto bail;
    }

        /* add the file list to the Apple Event */
    err = AEPutParamDesc(&theAEvent, keyDirectObject,
          &targetListDesc);
    if (err != noErr) goto bail;

        /* send the event to the Finder */
    err = AESend(&theAEvent, &theReply, kAENoReply,
        kAENormalPriority, kAEDefaultTimeout, NULL, NULL);

        /* clean up and leave */
bail:
    if (targetAlias != NULL) DisposeHandle((Handle) targetAlias);
    AEDisposeDesc(&targetListDesc);
    AEDisposeDesc(&theAEvent);
    AEDisposeDesc(&fndrAddress);
    AEDisposeDesc(&theReply);
    return err;
}
#endif

// Execute the command.
// 1. If HTML file, call stViewHTMLFile.
// 2. If there's an extension and an associated
//    application, run the app + document.
// 3. If the application exists, run the application.
bool AxFileUtilities::ExecuteFile(const wxString& command, const wxString& specificErrorMessage)
{
    wxString path, file, ext;

    // Don't be misled by non-extension dots in command paths
    SplitPath(command, & path, & file, & ext);

    wxString extLower(ext.Lower());
    if (extLower == wxT("html") || 
        extLower == wxT("htm") || 
        extLower == wxT("sgml"))
    {
        wxLaunchDefaultBrowser(command);
        return true;
    }
#ifdef __WXMAC__
    else
    {
        wxFileName fn(command);
        fn.Normalize(wxPATH_NORM_DOTS|wxPATH_NORM_TILDE|wxPATH_NORM_ABSOLUTE);
        wxString cmd = fn.GetFullPath();
        wxSplitPath(command, & path, & file, & ext);

        if (ext.IsEmpty() || ext == wxT(".app"))
        {
              if (cmd.GetChar(0) != wxT('"'))
                  cmd = wxT("\"") + cmd + wxT("\"");
              return (wxExecute(cmd) != 0);
        }

        FSSpec fsSpec;
        wxMacFilename2FSSpec(command, &fsSpec);

        return ( FinderLaunch(1 /*one file*/, &fsSpec) == noErr ) ;
    }
#else
    else
    {
        bool isApp = false;
#ifdef __WXMSW__
        if (command.Lower().Find(wxT(".exe")) != wxNOT_FOUND ||
            command.Lower().Find(wxT(".com")) != wxNOT_FOUND ||
            command.Lower().Find(wxT(".cmd")) != wxNOT_FOUND ||
            command.Lower().Find(wxT(".bat")) != wxNOT_FOUND)
            isApp = true;
#endif
        if (ext.IsEmpty())
            isApp = true;

        if (isApp)
        {
#if 1
            long pid = wxExecute(command);

#ifdef __WXMSW__
            return (pid != 0);
#else
            // Unfortunately this still doesn't work on Linux: a valid PID
            // is still returned.
            bool success = wxProcess::Exists(pid);
            //wxPrintf(wxT("Process success: %d\n"), (int) success);
            return success;
#endif

#else
            wxString command2 = actualCommand ;
            if (command2.GetChar(0) != wxT('"'))
                command2 = wxT("\"") + actualCommand + wxT("\"");

            if (!arguments.IsEmpty())
            {
                command2 += wxT(" ");
                if (arguments.GetChar(0) != wxT('"'))
                    command2 += (wxT("\"") + arguments + wxT("\""));
                else
                    command2 += arguments;
            }
            //wxPrintf(wxT("Executing: %s\n"), command2.c_str());

            return (wxExecute(command2) != 0);
#endif
        }

        wxString msg;
        wxString errMsg = specificErrorMessage;
        if (errMsg.IsEmpty())
        {
#ifdef __WXMSW__
            errMsg = _("Please install a suitable application.");
#else
            errMsg = _("Please install a suitable application and/or edit your MIME types.");
#endif
        }

        if (!ext.IsEmpty())
        {
            wxFileType *ft = wxTheMimeTypesManager->GetFileTypeFromExtension(ext);
            if ( !ft )
            {
                wxString msg;
                msg.Printf(_("Could not determine the application for extension %s.\n%s"), (const wxChar*) ext, (const wxChar*) errMsg);
                wxMessageBox(msg, wxT("Run command"), wxOK|wxICON_EXCLAMATION);
                return false;
            }
            
            wxString cmd;
            wxString params = command;
#ifdef __UNIX__
            if (params.GetChar(0) != wxT('"'))
                params = wxT("\"") + params + wxT("\"");
#endif
            bool ok = ft->GetOpenCommand(&cmd,
                wxFileType::MessageParameters(params));
#ifdef __UNIX__
            // Just in case we accidentally have double-quotes
            cmd.Replace(wxT("\"\""), wxT("\""));
#endif
            delete ft;
            
            if (!ok)
            {
                // TODO: some kind of configuration dialog here.
                wxString msg;
                msg.Printf(wxT("Could not determine the command for running %s.\n%s"), (const wxChar*) command, (const wxChar*) errMsg);

                wxMessageBox(msg, wxT("Run command"), wxOK|wxICON_EXCLAMATION);
                return false;
            }

            wxLogDebug(cmd);
            ok = (wxExecute(cmd, false) != 0);
            return ok;
        }
        return false;
    }
#endif
    return true;
}

// Convert to a valid filename (assumes no path separators in filename)
wxString AxFileUtilities::MakeValidFilename(const wxString& filename, bool allowSpaces)
{
    wxString str(filename);
    str.Replace(wxT("\\"), wxEmptyString);
    str.Replace(wxT("/"), wxEmptyString);
    str.Replace(wxT("/"), wxEmptyString);
    if (!allowSpaces)
        str.Replace(wxT(" "), wxEmptyString);
    str.Replace(wxT("\""), wxEmptyString);
    str.Replace(wxT("'"), wxEmptyString);
    str.Replace(wxT(","), wxEmptyString);
    str.Replace(wxT("("), wxEmptyString);
    str.Replace(wxT(")"), wxEmptyString);
    str.Replace(wxT("*"), wxEmptyString);
    str.Replace(wxT("?"), wxEmptyString);
    str.Replace(wxT("%"), wxEmptyString);
    str.Replace(wxT("$"), wxEmptyString);
    str.Replace(wxT("!"), wxEmptyString);
    str.Replace(wxT("{"), wxEmptyString);
    str.Replace(wxT("}"), wxEmptyString);
    str.Replace(wxT("["), wxEmptyString);
    str.Replace(wxT("]"), wxEmptyString);
    str.Replace(wxT("+"), wxEmptyString);
    str.Replace(wxT("~"), wxEmptyString);
    str.Replace(wxT("#"), wxEmptyString);
    str.Replace(wxT(":"), wxEmptyString);
    str.Replace(wxT(";"), wxEmptyString);
    str.Replace(wxT("@"), wxEmptyString);
    str.Replace(wxT("<"), wxEmptyString);
    str.Replace(wxT(">"), wxEmptyString);
    str.Replace(wxT("|"), wxEmptyString);
    str.Replace(wxT("&"), wxEmptyString);
    str.Replace(wxT("^"), wxEmptyString);
    str.Replace(wxT("`"), wxEmptyString);
    str.Replace(wxT("^"), wxEmptyString);

    return str;
}

/*
 * Archive utilities
 */

// Get the zip entry for the given filename
wxZipEntry* AxArchiveUtilities::FindEntry(wxZipInputStream& stream, const wxString& filename)
{
    wxZipEntry* zipEntry = NULL;
    while (true)
    {
        if (zipEntry)
        {
            delete zipEntry;
        }
        zipEntry = stream.GetNextEntry();
        if (!zipEntry)
            break;
        else
        {
            wxString zipFile = zipEntry->GetInternalName();
            if (zipFile == filename || zipFile.Matches(filename))
                return zipEntry;
        }
    }
    return NULL;
}

// Copy a fixed size of data from an input stream to an output stream,
// using a buffer to speed it up.
bool AxArchiveUtilities::BufferedCopy(wxInputStream& inStream, wxOutputStream& outStream, size_t size)
{
    return AxFileUtilities::BufferedCopy(inStream, outStream, size);
}

// Copy stream to new file
bool AxArchiveUtilities::CopyStreamToFile(wxInputStream& stream, const wxString& filename, size_t size)
{
    return AxFileUtilities::CopyStreamToFile(stream, filename, size);
}

// Copy entry to new file
bool AxArchiveUtilities::CopyEntryToFile(wxZipInputStream& stream, const wxString& filename, wxZipEntry* entry)
{
    stream.OpenEntry(*entry);

    return CopyStreamToFile(stream, filename, stream.GetSize());
}

// Unarchive to the given folder
bool AxArchiveUtilities::UnzipTo(const wxString& archive, const wxString& destinationFolder, wxProgressDialog* progressDialog)
{
    int i = 0;
    wxFFileInputStream inputStr(archive);
    if (inputStr.Ok())
    {
        wxZipInputStream zipStr(inputStr);

        wxZipEntry* zipEntry = NULL;
        while (true)
        {
            if (zipEntry)
            {
                delete zipEntry;
            }
            zipEntry = zipStr.GetNextEntry();
            if (!zipEntry)
                break;
            else if (!zipEntry->IsDir())
            {
                wxString zipFile = zipEntry->GetInternalName();

                if (progressDialog)
                {
                    wxString msg(wxString::Format(_("Writing %s"), zipFile.c_str()));
                    if (!progressDialog->Update(i, msg))
                    {
                        // Aborting on user request
                        delete zipEntry;
                        return false;
                    }
                }

                // Extract to destination

                wxString destFilename = destinationFolder + wxFILE_SEP_PATH + zipFile;
                zipStr.OpenEntry(*zipEntry);

                // TODO: for Unix, will need to make binaries executable.
                // May also be permission issues.

                wxString path = wxPathOnly(destFilename);
                if (!wxDirExists(path))
                    AxFileUtilities::MkDirRecursively(path);

                if (!CopyStreamToFile(zipStr, destFilename, zipStr.GetSize()))
                {
                    delete zipEntry;
                    return false;
                }

            }
            i ++;
        }
    }

    return true;
}

// Get the number of entries in the zip file
int AxArchiveUtilities::GetZipEntryCount(const wxString& filename)
{
    wxFFileInputStream inputStr(filename);
    if (inputStr.Ok())
    {
        wxZipInputStream zipStr(inputStr);
        return zipStr.GetTotalEntries();
    }
    else
        return 0;
}

// Unarchive to the given folder, optionally notifying the progress dialog
bool AxArchiveUtilities::GUnzipTo(const wxString& archive, const wxString& destinationFile)
{
    wxFFileInputStream inputStream(archive);
    wxZlibInputStream zStream(inputStream);
    wxFFileOutputStream outputStream(destinationFile);

    return AxFileUtilities::BufferedCopy(zStream, outputStream);
}

// Unarchive to the given folder, optionally notifying the progress dialog
bool AxArchiveUtilities::UntarTo(const wxString& archive, const wxString& destinationFolder, wxProgressDialog* progressDialog)
{
    int i = 0;
    wxFFileInputStream inputStr(archive);
    if (inputStr.Ok())
    {
        wxTarInputStream tarStr(inputStr);

        wxTarEntry* tarEntry = NULL;
        while (true)
        {
            if (tarEntry)
            {
                delete tarEntry;
            }
            tarEntry = tarStr.GetNextEntry();
            if (!tarEntry)
                break;
            else
            {
                if (tarEntry->GetTypeFlag() == wxTAR_REGTYPE)
                {
                    wxString tarFile = tarEntry->GetInternalName();


                    if (progressDialog)
                    {
                        wxString msg(wxString::Format(_("Writing %s"), tarFile.c_str()));
                        if (!progressDialog->Update(i, msg))
                        {
                            // Aborting on user request
                            delete tarEntry;
                            return false;
                        }
                    }

                    // Extract to destination

                    wxString destFilename = destinationFolder + wxFILE_SEP_PATH + tarFile;
                    tarStr.OpenEntry(*tarEntry);

                    wxString path = wxPathOnly(destFilename);
                    if (!wxDirExists(path))
                        AxFileUtilities::MkDirRecursively(path);

                    if (!CopyStreamToFile(tarStr, destFilename, tarStr.GetSize()))
                    {
                        delete tarEntry;
                        return false;
                    }
                    else
                    {
#ifdef __UNIX__
                        AxFileUtilities::Chmod(destFilename, tarEntry->GetMode());
#endif
                    }
                }

                i ++;
            }
        }
    }

    return true;
}

// Copy to the given folder, optionally notifying the progress dialog.
// relativeTo is the folder to be substracted from each file to determine the relative path for the destination file.
bool AxArchiveUtilities::CopyTo(const wxString& relativeTo, const wxArrayString& files, const wxString& destinationFolder, wxProgressDialog* progressDialog)
{
    size_t i;
    for (i = 0; i < files.GetCount(); i++)
    {
        wxString sourceFile = files[i];

        wxFileName fname(sourceFile);
        fname.MakeRelativeTo(relativeTo);
        wxString relativeFile = fname.GetFullPath();

        wxString destFilename = destinationFolder + wxFILE_SEP_PATH + relativeFile;

        wxString path = wxPathOnly(destFilename);

        if (progressDialog)
        {
            wxString msg(wxString::Format(_("Writing %s"), relativeFile.c_str()));
            if (!progressDialog->Update(i, msg))
            {
                // Aborting on user request
                return false;
            }
        }

        if (!wxDirExists(path))
            AxFileUtilities::MkDirRecursively(path);

        if (!wxCopyFile(sourceFile, destFilename))
            return false;
    }

    return true;
}

// Get a reasonable themed border style
wxBorder AxUIUtilities::GetSuitableThemedBorderStyle()
{
#if defined(__WXMSW__)
#if wxUSE_UXTHEME
    // If on Windows XP/Vista, use more appropriate colours.
    if (wxUxThemeEngine::GetIfActive())
    {
#if wxCHECK_VERSION(2,8,4)
        return wxBORDER_THEME;
#else
        return wxBORDER_SIMPLE;
#endif
    }
#endif
    return wxBORDER_SUNKEN;
#else
    return wxBORDER_SUNKEN;
#endif
}

// Choose a folder
void AxUIUtilities::ChooseFolder(wxTextCtrl* textCtrl, wxWindow* win)
{
    wxString msg = _("Choose a folder");
    wxWindow* parent = wxGetTopLevelParent(win);

    wxString originalPath = textCtrl->GetValue();

    wxDirDialog dirDlg(parent, msg, originalPath, wxDD_DEFAULT_STYLE);
    if (dirDlg.ShowModal() == wxID_OK)
    {
        wxString path = dirDlg.GetPath();
    
        textCtrl->SetValue(path);
    }
}

// Choose a file
void AxUIUtilities::ChooseFile(wxTextCtrl* textCtrl, const wxString& wildcard, const wxString& ext, wxWindow* win)
{
    wxString msg = _("Choose a file");
    wxWindow* parent = wxGetTopLevelParent(win);
    wxString filename(textCtrl->GetValue());

    wxString file(wxFileNameFromPath(filename));
    wxString path(wxPathOnly(filename));

    wxFileDialog fileDlg(parent, msg, path, file, wildcard);
    if (fileDlg.ShowModal() == wxID_OK)
    {
        path = fileDlg.GetPath();
        if (!ext.IsEmpty())
            AxFileUtilities::EnsureExtension(path, ext);

        textCtrl->SetValue(path);
    }
}

/// Find a child window by class
wxWindow* AxUIUtilities::FindWindowByClass(wxWindow* parent, wxClassInfo* info)
{
    wxWindowList::compatibility_iterator node = parent->GetChildren().GetFirst();
    while (node)
    {
        wxWindow *child = node->GetData();
        if (child->IsKindOf(info))
            return child;

        node = node->GetNext();
    }
    return NULL;
}

// wxChildWindowFromPoint
//
// This is a routine derived from wxFindWindowAtPoint (the unexported variant thereof).
// It finds the child window of a window passed as a parameter, which contains the point
// passed in. It fill find the deepest such child window. If no child window contains the
// point passed in, it will return a pointer to the window itself, unless that also
// does not contain the point in question, in which case it will return NULL.
// The "hidden" flag determines whether or not it should examine windows which are
// hidden (i.e. for which IsShown() is false). The routine can be passed a depth
// to recurse to, negative depth being infinite recursion, zero meaning do not look at
// child windows, and positive meaning a limited depth.
//
// Inputs:
//   wxWindow * win       The window to examine, plus children thereof
//   const wxPoint& pt    The coordinate, in screen coords
//   bool hidden          Flag, true to examine hidden windows as well (defaults to true)
//   int depth            Depth of recursion; zero to not recurse, positive for that number
//                        of levels, negative for infinite recursion (defaults to 1)
// Returns:
//   wxWindow *           Pointer to the deepest window searched containing pt, or NULL

wxWindow* AxUIUtilities::ChildWindowFromPoint(wxWindow* win, const wxPoint& pt, bool hidden /* =true */, int depth /* =-1 */)
{
    bool shown = win->IsShown();

    if (win->IsKindOf(CLASSINFO(wxMDIChildFrame)))
    {
        wxWindow * pParent=win;
        // Look for a wxMDIParentFrame
        while ((pParent = /*assignment*/ pParent->GetParent()) && !(pParent->IsKindOf(CLASSINFO(wxMDIParentFrame)))) {}
        if (pParent)
        {
            // If this isn't the active child, then treat it as if it was hidden.
            if (((wxMDIParentFrame *)pParent)->GetActiveChild() != win)
                shown=FALSE;
        }
    }

    if (!(hidden || shown))
        return NULL;

    // Hack for wxNotebook case: at least in wxGTK, all pages
    // claim to be shown, so we must only deal with the selected one.
#if wxUSE_NOTEBOOK
    // If depth is >0, we want to descend as we have more searching to do. If depth
    // is negative, that means descend to infinite depth. So don't descend if depth
    // is exactly zero
    if (depth && (win->IsKindOf(CLASSINFO(wxNotebook))))
    {
      wxNotebook* nb = (wxNotebook*) win;
      int sel = nb->GetSelection();
      if (sel >= 0)
      {
        wxWindow* child = nb->GetPage(sel);
        wxWindow* foundWin = ChildWindowFromPoint(child, pt, hidden, (depth<0)?depth:depth-1);
        if (foundWin)
           return foundWin;
      }
    }
#endif

    // If depth is >0, we want to descend as we have more searching to do. If depth
    // is negative, that means descend to infinite depth. So don't descend if depth
    // is exactly zero
    if (depth)
    {
        wxWindowList::compatibility_iterator node = win->GetChildren().GetLast();
        while (node)
        {
            wxWindow* child = node->GetData();
            wxWindow* foundWin = ChildWindowFromPoint(child, pt, hidden, (depth<0)?depth:depth-1);
            if (foundWin)
              return foundWin;
            node = node->GetPrevious();
        }
    }

    wxPoint pos = win->GetPosition();
    wxSize sz = win->GetSize();
    if (win->GetParent() && !win->IsTopLevel()) // Do not translate coords in TLWs as they are already screen coords
    {
        pos = win->GetParent()->ClientToScreen(pos);
    }

    wxRect rect(pos, sz);
    if (rect.Contains(pt))
        return win;
    else
        return NULL;
}

// The same routine again but without the window hint

wxWindow* AxUIUtilities::ChildWindowFromPoint(const wxPoint& pt, bool hidden /* =true */, int depth /* =-1 */)
{
    // Go backwards through the list since windows
    // on top are likely to have been appended most
    // recently.
    wxWindowList::compatibility_iterator node = wxTopLevelWindows.GetLast();
    while (node)
    {
        wxWindow* win = node->GetData();
        wxWindow* found = ChildWindowFromPoint(win, pt, hidden, depth);
        if (found)
            return found;
        node = node->GetPrevious();
    }
    return NULL;
}


// Calculate a suitable scale for an image to fit into the given space
wxSize AxCalculateScaledImageSize(const wxSize& maxSize, const wxSize& imageSize)
{
    double scaleX = ((double) maxSize.x)/((double) imageSize.x);
    double scaleY = ((double) maxSize.y)/((double) imageSize.y);
    
    if (scaleX < 1.0 || scaleY < 1.0)
    {
        double scale = wxMin(scaleX, scaleY);
        int newWidth = (int) (scale * imageSize.x);
        int newHeight = (int) (scale * imageSize.y);
        return wxSize(newWidth, newHeight);
    }
    else
        return imageSize;
}

// Remove bad characters from a filename
bool AxRemoveNonFilenameCharacters(wxString& str)
{
    str.Replace(wxT(" "), wxEmptyString);
    str.Replace(wxT(","), wxEmptyString);
    str.Replace(wxT(":"), wxEmptyString);
    str.Replace(wxT("?"), wxEmptyString);
    str.Replace(wxT("!"), wxEmptyString);
    str.Replace(wxT("*"), wxEmptyString);
    str.Replace(wxT("\""), wxEmptyString);
    str.Replace(wxT("\'"), wxEmptyString);
    str.Replace(wxT("/"), wxEmptyString);
    str.Replace(wxT("\\"), wxEmptyString);
    str.Replace(wxT("("), wxEmptyString);
    str.Replace(wxT(")"), wxEmptyString);

    return true;
}

static const struct AxKeyName
{
    int code;
    const wxChar *name;
} AxKeyNames[] =
{
    { WXK_DELETE, wxTRANSLATE("delete") },
    { WXK_BACK, wxTRANSLATE("back") },
    { WXK_INSERT, wxTRANSLATE("insert") },
    { WXK_RETURN, wxTRANSLATE("return") },
    { WXK_PAGEUP, wxTRANSLATE("page up") },
    { WXK_PAGEDOWN, wxTRANSLATE("page down") },
    { WXK_LEFT, wxTRANSLATE("left") },
    { WXK_RIGHT, wxTRANSLATE("right") },
    { WXK_UP, wxTRANSLATE("up") },
    { WXK_DOWN, wxTRANSLATE("down") },
    { WXK_HOME, wxTRANSLATE("home") },
    { WXK_END, wxTRANSLATE("end") },
    { WXK_SPACE, wxTRANSLATE("space") },
    { WXK_TAB, wxTRANSLATE("tab") },
    { WXK_ESCAPE, wxTRANSLATE("escape") },
    { WXK_CANCEL, wxTRANSLATE("cancel") },
    { WXK_CLEAR, wxTRANSLATE("clear") },
    { WXK_MENU, wxTRANSLATE("menu") },
    { WXK_PAUSE, wxTRANSLATE("pause") },
    { WXK_CAPITAL, wxTRANSLATE("capital") },
    { WXK_SELECT, wxTRANSLATE("select") },
    { WXK_PRINT, wxTRANSLATE("print") },
    { WXK_EXECUTE, wxTRANSLATE("execute") },
    { WXK_SNAPSHOT, wxTRANSLATE("snapshot") },
    { WXK_HELP, wxTRANSLATE("help") },
    { WXK_ADD, wxTRANSLATE("add") },
    { WXK_SEPARATOR, wxTRANSLATE("separator") },
    { WXK_SUBTRACT, wxTRANSLATE("subtract") },
    { WXK_DECIMAL, wxTRANSLATE("decimal") },
    { WXK_DIVIDE, wxTRANSLATE("divide") },
    { WXK_NUMLOCK, wxTRANSLATE("num lock") },
    { WXK_SCROLL, wxTRANSLATE("scroll lock") },
    { WXK_PAGEUP, wxTRANSLATE("page up") },
    { WXK_PAGEDOWN, wxTRANSLATE("page down") },
    { WXK_NUMPAD_SPACE, wxTRANSLATE("space") },
    { WXK_NUMPAD_TAB, wxTRANSLATE("tab") },
    { WXK_NUMPAD_ENTER, wxTRANSLATE("return") },
    { WXK_NUMPAD_HOME, wxTRANSLATE("home") },
    { WXK_NUMPAD_LEFT, wxTRANSLATE("leftt") },
    { WXK_NUMPAD_UP, wxTRANSLATE("up") },
    { WXK_NUMPAD_RIGHT, wxTRANSLATE("right") },
    { WXK_NUMPAD_DOWN, wxTRANSLATE("down") },
    { WXK_NUMPAD_PAGEUP, wxTRANSLATE("page up") },
    { WXK_NUMPAD_PAGEDOWN, wxTRANSLATE("page down") },
    { WXK_NUMPAD_END, wxTRANSLATE("end") },
    { WXK_NUMPAD_BEGIN, wxTRANSLATE("begin") },
    { WXK_NUMPAD_INSERT, wxTRANSLATE("insert") },
    { WXK_NUMPAD_DELETE, wxTRANSLATE("delete") },
    { WXK_NUMPAD_EQUAL, wxTRANSLATE("equals") },
    { WXK_NUMPAD_MULTIPLY, wxTRANSLATE("multiply") },
    { WXK_NUMPAD_ADD, wxTRANSLATE("add") },
    { WXK_NUMPAD_SEPARATOR, wxTRANSLATE("separator") },
    { WXK_NUMPAD_SUBTRACT, wxTRANSLATE("subtract") },
    { WXK_NUMPAD_DECIMAL, wxTRANSLATE("decimal") },
    { WXK_NUMPAD_DIVIDE, wxTRANSLATE("divide") },
    { WXK_WINDOWS_LEFT, wxTRANSLATE("left") },
    { WXK_WINDOWS_RIGHT, wxTRANSLATE("right") },
    { WXK_WINDOWS_MENU, wxTRANSLATE("menu") },
    { WXK_COMMAND, wxTRANSLATE("command") },
    { WXK_F1, wxTRANSLATE("function 1") },
    { WXK_F2, wxTRANSLATE("function 2") },
    { WXK_F3, wxTRANSLATE("function 3") },
    { WXK_F4, wxTRANSLATE("function 4") },
    { WXK_F5, wxTRANSLATE("function 5") },
    { WXK_F6, wxTRANSLATE("function 6") },
    { WXK_F7, wxTRANSLATE("function 7") },
    { WXK_F8, wxTRANSLATE("function 8") },
    { WXK_F9, wxTRANSLATE("function 9") },
    { WXK_F10, wxTRANSLATE("function 10") },
    { WXK_F11, wxTRANSLATE("function 11") },
    { WXK_F12, wxTRANSLATE("function 12") },
    { WXK_F13, wxTRANSLATE("function 13") },
    { WXK_F14, wxTRANSLATE("function 14") },
    { WXK_F15, wxTRANSLATE("function 15") },

    { '|', wxTRANSLATE("pipe") },
    { '\\', wxTRANSLATE("backslash") },
    { '`', wxTRANSLATE("backquote") },
    { '!',  wxTRANSLATE("exclamation") },
    { '"',  wxTRANSLATE("double quotes") },
    { '',  wxTRANSLATE("pound") },
    { '$',  wxTRANSLATE("dollar") },
    { '%',  wxTRANSLATE("percent") },
    { '^',  wxTRANSLATE("circumflex") },
    { '&',  wxTRANSLATE("ampersand") },
    { '*',  wxTRANSLATE("asterisk") },
    { '(',  wxTRANSLATE("open parenthesis") },
    { ')',  wxTRANSLATE("close parenthesis") },
    { '_',  wxTRANSLATE("underscore") },
    { '-',  wxTRANSLATE("hyphen") },
    { '+',  wxTRANSLATE("plus") },
    { '=',  wxTRANSLATE("equals") },
    { '{',  wxTRANSLATE("open curly bracket") },
    { '}',  wxTRANSLATE("close curly bracket") },
    { '[',  wxTRANSLATE("open square bracket") },
    { ']',  wxTRANSLATE("close square bracket") },
    { '@',  wxTRANSLATE("at") },
    { '\'',  wxTRANSLATE("single quote") },
    { '~',  wxTRANSLATE("tilde") },
    { '#',  wxTRANSLATE("hash") },
    { ',',  wxTRANSLATE("comma") },
    { '.',  wxTRANSLATE("stop") },
    { '<',  wxTRANSLATE("less than") },
    { '>',  wxTRANSLATE("more than") },
    { '?',  wxTRANSLATE("question mark") },
    { '/',  wxTRANSLATE("forward slash") }
};

// Gets suitable text for a given keycode
wxString AxGetSpeakableTextForKeyCode(int keycode)
{
    size_t i;
    for (i = 0; i < WXSIZEOF(AxKeyNames); i++)
    {
        if (AxKeyNames[i].code == keycode)
            return AxKeyNames[i].name;
    }
    wxString text;
    text += (wxChar) keycode;
    return text;
}
